#!/bin/bash

# otupipe.bash: cluster 454 reads into OTUs.
# (Not designed for or tested with Illumina reads).
# v9, tested with usearch v4.2.66.

# Usage
# -----
#    otupipe.bash reads.fasta outputdir [existing_otus.fasta]

# Author
# ------
# Robert Edgar
# robert@drive5.com
#
# Home page for this script:
#    http://drive5.com/otupipe
#

# Output files.
# -------------
# Output file names are hard-coded into this script.
# The name of the directory where the output files are created is
# given as a command-line option. If the directory does not exist,
# it is created. If this script is run on a cluster, then you
# should make sure the output directory name is unique in order
# to avoid collisions.
#
#       Filename   Description
#       --------   -----------
#    chimeras.fa   Centroid sequences for chimeras clustered at 97%.
#        otus.fa   Centroid sequence for each OTU.
#     readmap.uc   Uclust file assigning reads to OTUs or chimeras.

# Required parameters.
# -------------------
# These two environment variables MUST be set:
#	$u Path to usearch binary.
#	$UCHIME_REFDB Path to FASTA file containing uchime reference database.
#	  Suggested database is gold.fa.

# Defaulted parameters.
# --------------------
# These can be changed by setting the corresponding environment variables.
# E.g., to set MINSIZE to 8, do this before running the script:
#    export MINSIZE=8

DEFAULT_MINSIZE=4      # Discard clusters < MINSIZE in error correction round
DEFAULT_PCTID_ERR=97   # Id threshold for error correction round
DEFAULT_PCTID_OTU=97   # Id threshold for OTU clustering
DEFAULT_PCTID_BIN=97   # Id threshold for read binning
DEFAULT_ABSKEW=2       # Abundance skew for de novo chimera filtering
DEFAULT_OTUTMPDIR=./otutmp.$$.$RANDOM	# Name for directory to store temporary files
DEFAULT_PREFIX=OTU_	   # Prefix for OTU label, default is OTU_ giving OTU_1, OTU_2 ...
DEFAULT_CHIMPREFIX=CHIMERA_  # Prefix for chimera label, default is CHIM_ giving CHIM_1, OTU_2 ...

input=$1
outdir=$2
existing_otus=$3

if [ x$outdir == x ] ; then
	echo >> /dev/stderr
	echo "Usage: $0 reads.fasta outputdir [existing_otus.fasta]" >> /dev/stderr
	echo >> /dev/stderr
	exit 1
fi

echo >> /dev/stderr
echo otupipe v9 >> /dev/stderr
echo http://drive5.com/otupipe >> /dev/stderr
echo >> /dev/stderr

if [ x$MINSIZE == x ] ; then
	MINSIZE=$DEFAULT_MINSIZE
fi

if [ x$PCTID_ERR == x ] ; then
	PCTID_ERR=$DEFAULT_PCTID_ERR
fi

if [ x$PCTID_OTU == x ] ; then
	PCTID_OTU=$DEFAULT_PCTID_OTU
fi

if [ x$PCTID_BIN == x ] ; then
	PCTID_BIN=$DEFAULT_PCTID_BIN
fi

if [ x$ABSKEW == x ] ; then
	ABSKEW=$DEFAULT_ABSKEW
fi

if [ x$UCHIME_REFDB == x ] ; then
	echo "\$UCHIME_REFDB not set, use e.g. export UCHIME_REFDB=/public/db/gold.fa" >> /dev/stderr
	exit 1
fi

if [ x$u == x ] ; then
	echo "\$u not set, use e.g. export u=/usr/bin/usearch4.2.66_i86linux32" >> /dev/stderr
	exit 1
fi

if [ x$OTUTMPDIR == x ] ; then
	OTUTMPDIR=$DEFAULT_OTUTMPDIR
fi

if [ x$PREFIX == x ] ; then
	PREFIX=$DEFAULT_PREFIX
fi

if [ x$CHIMPREFIX == x ] ; then
	CHIMPREFIX=$DEFAULT_CHIMPREFIX
fi

mkdir -p $outdir
if [ ! -d $outdir ]; then
	echo "Failed to make outputdir or doesn't exist: " $outdir >> /dev/stderr
	exit 1
fi

t=$OTUTMPDIR
mkdir -p $t

if [ ! -d $t ] ; then
	echo "Failed to make tmp dir or doesn't exist: " $t >> /dev/stderr
	exit 1
fi

currdir=`pwd`

echo ""
echo Input=$input
echo MINSIZE=$MINSIZE
echo PCTID_ERR=$PCTID_ERR
echo PCTID_OTU=$PCTID_OTU
echo PCTID_BIN=$PCTID_BIN
echo UCHIME_REFDB=$UCHIME_REFDB
echo ABSKEW=$ABSKEW
echo OTUTMPDIR=$OTUTMPDIR
echo PREFIX=$PREFIX
echo CHIMPREFIX=$CHIMPREFIX
echo

#=========
# Pipeline
#=========

nseqs_input=`grep ">" $input | wc -l`
echo $nseqs_input reads

echo
echo =========================================================================
echo "Sort by length, $nseqs_input sequences"
echo =========================================================================
$u --sort $input --output $t/input.sortlen --log $t/sortlen.log
if [ $? != 0 ] ; then
	exit 1
fi
ls -l $t/input.sortlen

echo
echo =========================================================================
echo "Dereplicate exact sub-sequences"
echo =========================================================================
$u --cluster $t/input.sortlen --derep_subseq --uc $t/derep.uc --seedsout $t/derep.fa --minlen 64 --w 64 --slots 16769023 --log $t/derep.log --sizeout --maxrejects 64
if [ $? != 0 ] ; then
	exit 1
fi
ls -l derep.fa
nseqs_derep=`grep ">" $t/derep.fa | wc -l`
echo $nseqs_derep seqs after derep

echo
echo =========================================================================
echo "Sort by abundance, output to derep.fas."
echo =========================================================================
$u --sortsize $t/derep.fa --output $t/derep.fas --log $t/sortsize1.log --sizein --sizeout
if [ $? != 0 ] ; then
	exit 1
fi
ls -l $t/derep.fas

echo
echo =========================================================================
echo "Cluster for err. corr. at $PCTID_ERR pct, output consensus to cons.fa."
echo =========================================================================
$u --cluster $t/derep.fas --consout $t/cons.fa --id 0.$PCTID_ERR --log $t/cluster_err.log --w 20 --slots 16769023 --maxrejects 64 --sizein --sizeout --usersort
if [ $? != 0 ] ; then
	exit 1
fi
ls -l $t/cons.fa
nseqs_cons=`grep ">" $t/cons.fa | wc -l`
echo $nseqs_cons consensus seqs

echo
echo =========================================================================
echo "Chimera filter de novo, output tmp1.fa."
echo =========================================================================
$u --uchime $t/cons.fa --chimeras $t/chimeras_denovo.fa --nonchimeras $t/tmp1.fa --uchimeout $t/denovo.uchime --log $t/uchimedn.log --abskew $ABSKEW
if [ $? != 0 ] ; then
	exit 1
fi
ls -l $t/tmp1.fa
nseqs_dn=`grep ">" $t/tmp1.fa | wc -l`
echo $nseqs_dn seqs after de novo chimera filter

echo
echo =========================================================================
echo "Chimera filter ref. db., output tmp2.fa"
echo =========================================================================
$u --uchime $t/tmp1.fa --db $UCHIME_REFDB --rev --uchimeout $t/refdb.uchime --chimeras $t/chimeras_ref.fa --nonchimeras $t/tmp2.fa --log $t/uchimedb.log
if [ $? != 0 ] ; then
	exit 1
fi
ls -l $t/tmp2.fa
nseqs_db=`grep ">" $t/tmp2.fa | wc -l`
echo $nseqs_db seqs after ref db chimera filter

cat $t/chimeras_denovo.fa $t/chimeras_ref.fa > $t/ch.fa

echo
echo =========================================================================
echo "Sort chimeric sequences by abundance, output tmp2c.fa"
echo =========================================================================

# Bug in usearch v4.2.66, crashes if input file is empty
if [ -s $t/ch.fa ] ; then
	$u --sortsize $t/ch.fa --output $t/tmp2c.fas --log $t/sort2c.log
else
	rm -f $t/tmp2c.fas
	touch $t/tmp2c.fas
fi

if [ $? != 0 ] ; then
	exit 1
fi
ls -l $t/tmp2c.fas

echo
echo =========================================================================
echo "Cluster chimeras at $PCTID_OTU, output in tmp3c.fa"
echo =========================================================================
$u --cluster $t/tmp2c.fas --id 0.$PCTID_OTU --uc $t/otu1.uc --seedsout $t/tmp3c.fa --w 20 --slots 16769023 --log $t/cluster2c.log --sizein --sizeout --maxrejects 64 --usersort

echo
echo =========================================================================
echo "Assign sequential numbers to chimeric OTUs"
echo =========================================================================
fasta_number.py $t/tmp3c.fa ${CHIMPREFIX} > $outdir/chimeras.fa
if [ $? != 0 ] ; then
	exit 1
fi
ls -l $outdir/chimeras.fa
nchimotus=`grep ">" $outdir/chimeras.fa | wc -l`
echo $nchimotus chimeras

echo
echo =========================================================================
echo "Sort chimera-filtered seqs by abundance, output to tmp2.fas."
echo "Discard clusters < MINSIZE=$MINSIZE."
echo =========================================================================
$u --sortsize $t/tmp2.fa --output $t/tmp2.fas --minsize $MINSIZE --log $t/sort2.log
if [ $? != 0 ] ; then
	exit 1
fi
ls -l $t/tmp2.fas

echo
echo =========================================================================
echo "Cluster at $PCTID_OTU to merge OTUs, output in tmp3.fa"
echo =========================================================================

if [ x$existing_otus == x ] ; then
	xcmd=
else
	xcmd="--db $existing_otus"
fi

$u --cluster $t/tmp2.fas --id 0.$PCTID_OTU --uc $t/otu1.uc --seedsout $t/tmp3.fa --w 20 --slots 16769023 --log $t/cluster2.log --sizein --sizeout --maxrejects 64 --usersort $xcmd
if [ $? != 0 ] ; then
	exit 1
fi
ls -l $t/tmp3.fa

echo
echo =========================================================================
echo "Assign sequential numbers to OTUs, output is otus.fa."
echo =========================================================================
fasta_number.py $t/tmp3.fa ${PREFIX} $existing_otus > $outdir/otus.fa
if [ $? != 0 ] ; then
	exit 1
fi
ls -l $outdir/otus.fa
nseqs_otu=`grep ">" $outdir/otus.fa | wc -l`
echo $nseqs_otu OTUs

cat $existing_otus $outdir/otus.fa $outdir/chimeras.fa > $t/db.fa

echo
echo =========================================================================
echo "Classify reads, output in read2otu.uc."
echo =========================================================================
$u --query $input --db $t/db.fa --log $t/bin.log --id 0.$PCTID_BIN --global --uc $outdir/readmap.uc
if [ $? != 0 ] ; then
	exit 1
fi
ls -l $outdir/readmap.uc

echo
echo =========================================================================
echo "Delete temp dir $t"
echo =========================================================================
rm -rf $t

echo =========================================================================
echo "Done. Output file in $outdir"
echo =========================================================================

ls -l $outdir
